from datetime import datetime
import os

from sphinx_timeline import __version__

project = "Sphinx Timeline"
author = "Chris Sewell"
copyright = f"{datetime.now().year}, Chris Sewell"
version = __version__

extensions = ["myst_parser", "sphinx_timeline"]

myst_enable_extensions = ["deflist"]

suppress_warnings = ["epub.unknown_project_files"]

# get environment variables
html_theme = os.environ.get("HTML_THEME", "furo")
html_static_path = ["_static"]
html_css_files = ["custom.css"]



"""Utilities for parsing and formatting dates and times."""
from __future__ import annotations

from datetime import date, datetime, time, timezone
import re

from dateutil.relativedelta import relativedelta

try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo  # noqa: F401


RE_ISO_TZ = re.compile(r"(?P<dt>.+)\((?P<tz>[^)]+)\)\s*$")


def to_datetime(value: str | date | datetime, default_tz=timezone.utc) -> datetime:
    """Convert to a timezone-aware datetime object."""
    final: datetime
    if isinstance(value, str):
        # find timezone name suffix in parentheses
        tz_match = RE_ISO_TZ.match(value)
        if tz_match:
            # parse datetime with timezone
            final = datetime.fromisoformat(tz_match.group("dt").strip())
            try:
                tz = zoneinfo.ZoneInfo(tz_match.group("tz"))
            except Exception as exc:
                raise ValueError(
                    f"Invalid timezone: {tz_match.group('tz')!r}"
                    # f"\navailable: {zoneinfo.available_timezones()!r}"
                ) from exc
            final = final.replace(tzinfo=tz)
        else:
            # parse datetime without timezone
            final = datetime.fromisoformat(value.strip())
    elif isinstance(value, date):
        final = datetime(value.year, value.month, value.day)
    elif not isinstance(value, datetime):
        raise TypeError(f"invalid type: {type(value)}")
    # if not timezone aware, assume UTC
    if final.tzinfo is None:
        final = final.replace(tzinfo=default_tz)
    return final


def parse_duration(value: str) -> relativedelta:
    """Parse a duration string."""
    if not isinstance(value, str):
        raise TypeError(f"not str: {type(value)}")
    delta = {}
    if year_match := re.search(r"(\d+)\s?y", value):
        delta["years"] = int(year_match.group(1))
    if month_match := re.search(r"(\d+)\s?mon", value):
        delta["months"] = int(month_match.group(1))
    if day_match := re.search(r"(\d+)\s?d", value):
        delta["days"] = int(day_match.group(1))
    if hour_match := re.search(r"(\d+)\s?h", value):
        delta["hours"] = int(hour_match.group(1))
    if minute_match := re.search(r"(\d+)\s?min", value):
        delta["minutes"] = int(minute_match.group(1))
    if second_match := re.search(r"(\d+)\s?s", value):
        delta["seconds"] = int(second_match.group(1))
    return relativedelta(**delta)


def fmt_daterange(
    start: datetime,
    duration: relativedelta | None = None,
    *,
    day_name=True,
    short_date=False,
    short_delim="/",
    abbr=True,
    clock12=True,
) -> str:
    """Format a datetime span.

    :param start: start datetime
    :param duration: duration of span
    :param day_name: include day name (e.g. "Monday")
    :param short_date: use short date format (e.g. "01/01/2021")
    :param short_delim: delimiter for short date format
    :param abbr: use abbreviated day/month name (e.g. "Mon")
    :param clock12: use AM/PM time format
    """
    day_name_code = ("%a " if abbr else "%A ") if day_name else ""
    day_code = f"%d{short_delim}" if short_date else "%D "
    month_code = f"%m{short_delim}" if short_date else ("%b " if abbr else "%B ")
    time_code = "%I:%M %p" if clock12 else "%H:%M"
    tz_code = " (%Z)" if start.tzinfo else ""

    start_fmt = f"{day_name_code}{day_code}{month_code}%Y, {time_code}{tz_code}"
    end_fmt = f"{day_name_code}{day_code}{month_code}%Y, {time_code}{tz_code}"

    if duration is None:
        end = start
    else:
        end = start + duration

    if start == end:
        if start.time() == time(0):
            # remove time from start format
            start_fmt = start_fmt.split(",")[0]
        return fmt_datetime(start, start_fmt)

    # remove timezone from start format
    start_fmt = start_fmt.replace(tz_code, "")

    if start.date() == end.date():
        # remove date from end format
        end_fmt = end_fmt.split(",")[1].lstrip()
        return f"{fmt_datetime(start, start_fmt)} - {fmt_datetime(end, end_fmt)}"

    # assume if the start time is 00:00:00, it was set without a time
    if start.time() == time(0) and start.time() == end.time():
        # remove time from start and end format
        start_fmt = start_fmt.split(",")[0]
        end_fmt = end_fmt.split(",")[0]

    if not short_date and start.year == end.year:
        start_fmt = start_fmt.replace(" %Y", "")
        if start.month == end.month:
            start_fmt = start_fmt.replace(f"{month_code}", "")

    return f"{fmt_datetime(start, start_fmt)} - {fmt_datetime(end, end_fmt)}"


def _ord_suffix(n: int):
    """Return ordinal suffix for a number."""
    return (
        "th" if 11 <= (n % 100) <= 13 else {1: "st", 2: "nd", 3: "rd"}.get(n % 10, "th")
    )


def fmt_datetime(dt: datetime, fmt: str) -> str:
    """Format a datetime.

    Extended strftime codes:
    - %D: day number with ordinal suffix (e.g. 1st, 2nd, 3rd, 4th, etc.)
    """
    # TODO deal with locales (for day names, month names, etc.)
    fmt = fmt.replace("%D", str(dt.day) + _ord_suffix(dt.day))
    return dt.strftime(fmt)


def fmt_delta(delta: relativedelta | None) -> str:
    """Format a relativedelta object."""
    if delta is None:
        return ""

    string = []

    if delta.years:
        string.append(f"{delta.years} Year{_s(delta.years)}")
    if delta.months:
        string.append(f"{delta.months} Month{_s(delta.months)}")
    if delta.days:
        string.append(f"{delta.days} Day{_s(delta.days)}")
    if delta.hours:
        string.append(f"{delta.hours} Hour{_s(delta.hours)}")
    if delta.minutes:
        string.append(f"{delta.minutes} Minute{_s(delta.minutes)}")
    if delta.seconds:
        string.append(f"{delta.seconds} Second{_s(delta.seconds)}")
    return " ".join(string)


def _s(n: int):
    """Return "s" if number is not 1."""
    return "s" if n != 1 else ""


class DtRangeStr:
    """A datetime range string format."""

    def __init__(self, start: datetime, duration: relativedelta | None = None) -> None:
        self._start = start
        self._duration = duration
        self._default = fmt_daterange(start, duration)

    def __str__(self) -> str:
        return self._default

    def __call__(self, **kwargs) -> str:
        return fmt_daterange(self._start, self._duration, **kwargs)



import json
from pathlib import Path

# from bs4 import BeautifulSoup
import pytest
from sphinx_pytest.plugin import CreateDoctree
import yaml

FIXTURE_PATH = Path(__file__).parent.joinpath("fixtures")


@pytest.mark.param_file(FIXTURE_PATH / "posttransform_html.txt")
def test_posttransform_html(file_params, sphinx_doctree: CreateDoctree):
    """Test AST output after post-transforms, when using the HTML builder."""
    sphinx_doctree.set_conf({"extensions": ["sphinx_timeline"]})
    sphinx_doctree.buildername = "html"
    example_data = [{"start": "2021-02-03", "name": "1st draft"}]
    sphinx_doctree.srcdir.joinpath("data.yaml").write_text(yaml.dump(example_data))
    sphinx_doctree.srcdir.joinpath("data.json").write_text(json.dumps(example_data))
    sphinx_doctree.srcdir.joinpath("template.txt").write_text(
        "{{dtrange}} - {{e.name}}"
    )
    result = sphinx_doctree(file_params.content)
    assert not result.warnings
    file_params.assert_expected(result.get_resolved_pformat(), rstrip_lines=True)


# @pytest.mark.param_file(FIXTURE_PATH / "build_html.txt")
# def test_doctree(file_params, sphinx_doctree: CreateDoctree):
#     """Test HTML build output."""
#     sphinx_doctree.set_conf({"extensions": ["sphinx_timeline"]})
#     sphinx_doctree.buildername = "html"
#     result = sphinx_doctree(file_params.content)
#     assert not result.warnings
#     html = BeautifulSoup(
#         Path(result.builder.outdir).joinpath("index.html").read_text(), "html.parser"
#     )
#     fig_html = str(html.select_one("figure.sphinx-subfigure"))
#     file_params.assert_expected(fig_html, rstrip_lines=True)



from __future__ import annotations

import csv
import hashlib
from importlib import resources
from io import StringIO
import json
from pathlib import Path
import re
from typing import Any, Literal, TextIO

from docutils import nodes
from docutils.parsers.rst import directives
from docutils.statemachine import StringList
import jinja2
from sphinx.application import Sphinx
from sphinx.util.docutils import SphinxDirective
import yaml

from sphinx_timeline import dtime
from sphinx_timeline import static as static_module


def setup(app: Sphinx) -> None:
    """Setup the extension."""
    from sphinx_timeline import __version__

    app.connect("builder-inited", add_html_assets)
    app.connect("html-page-context", load_html_assets)
    app.add_directive("timeline", TimelineDirective)
    app.add_node(
        TimelineDiv,
        html=(visit_tl_div, depart_tl_div),
        latex=(visit_depart_null, visit_depart_null),
        text=(visit_depart_null, visit_depart_null),
        man=(visit_depart_null, visit_depart_null),
        texinfo=(visit_depart_null, visit_depart_null),
    )

    return {
        "version": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }


def add_html_assets(app: Sphinx) -> None:
    """Add the HTML assets to the build directory."""
    if (not app.builder) or app.builder.format != "html":
        return
    # setup up new static path in output dir
    static_path = (Path(app.outdir) / "_sphinx_timeline_static").absolute()
    static_path.mkdir(exist_ok=True)
    for path in static_path.glob("**/*"):
        path.unlink()
    app.config.html_static_path.append(str(static_path))
    for resource in resources.contents(static_module):
        if not resource.endswith(".css") and not resource.endswith(".js"):
            continue
        # Read the content and hash it
        content = resources.read_text(static_module, resource)
        hash = hashlib.md5(content.encode("utf8")).hexdigest()
        # Write the file
        name, ext = resource.split(".", maxsplit=1)
        write_path = static_path / f"{name}.{hash}.{ext}"
        write_path.write_text(content, encoding="utf8")


def load_html_assets(app: Sphinx, pagename: str, *args, **kwargs) -> None:
    """Ensure the HTML assets are loaded in the page, if necessary."""
    if (not app.builder) or app.builder.format != "html":
        return
    if (not app.env) or not app.env.metadata.get(pagename, {}).get("timeline", False):
        return
    for resource in resources.contents(static_module):
        if not resource.endswith(".css") and not resource.endswith(".js"):
            continue
        # Read the content and hash it
        content = resources.read_text(static_module, resource)
        hash = hashlib.md5(content.encode("utf8")).hexdigest()
        # add the file to the context
        name, ext = resource.split(".", maxsplit=1)
        write_name = f"{name}.{hash}.{ext}"
        if ext == "css":
            app.add_css_file(write_name)
        if ext == "js":
            app.add_js_file(write_name)


class TimelineDiv(nodes.General, nodes.Element):
    """A div for a timeline."""

    def add_style(self, key, value):
        """Add a style to the div."""
        self.attributes.setdefault("styles", {})
        self["styles"][key] = value


def visit_depart_null(self, node: nodes.Element) -> None:
    """visit/depart passthrough"""


def visit_tl_div(self, node: nodes.Node):
    """visit tl_div"""
    attrs = {}
    if node.get("styles"):
        attrs["style"] = ";".join(
            (f"{key}: {val}" for key, val in node["styles"].items())
        )
    if node.get("dt"):
        attrs["data-dt"] = str(node["dt"])
    self.body.append(self.starttag(node, "div", CLASS="docutils", **attrs))


def depart_tl_div(self, node: nodes.Node):
    """depart tl_div"""
    self.body.append("</div>\n")


RE_BREAKLINE = re.compile(r"^\s*-{3,}\s*$")


class TimelineDirective(SphinxDirective):
    """A sphinx directive to create timelines."""

    has_content = True
    required_arguments = 0
    optional_arguments = 0
    option_spec = {
        "events": directives.path,
        "template": directives.path,
        "events-format": lambda val: directives.choice(val, ["yaml", "json", "csv"]),
        "max-items": directives.nonnegative_int,
        "reversed": directives.flag,
        "height": directives.length_or_unitless,
        "width-item": directives.length_or_percentage_or_unitless,
        "style": lambda val: directives.choice(val, ["default", "none"]),
        "class": directives.class_option,
        "class-item": directives.class_option,
    }

    def run(self) -> list[nodes.Element]:
        """Run the directive."""
        data: list[dict[str, Any]]
        template_lines: list[str]

        # get data
        if "events" in self.options:
            input_file = self.options["events"]
            # get file path relative to the document
            _, abs_path = self.env.relfn2path(input_file, self.env.docname)
            if not Path(abs_path).exists():
                raise self.error(f"'data' path does not exist: {abs_path}")
            # read file
            with Path(abs_path).open("r", encoding="utf8") as handle:
                try:
                    data = read_events(
                        handle, self.options.get("events-format", "yaml")
                    )
                except Exception as exc:
                    raise self.error(f"Error parsing data: {exc}")
            # add input file to dependencies
            self.env.note_dependency(input_file)

            template_lines = list(self.content)
        else:
            # split lines by first occurrence of line break
            data_lines = []
            template_lines = []
            in_template = False
            for line in self.content:
                if not in_template and RE_BREAKLINE.match(line):
                    in_template = True
                elif in_template:
                    template_lines.append(line)
                else:
                    data_lines.append(line)

                try:
                    data = read_events(
                        StringIO("\n".join(data_lines)),
                        self.options.get("events-format", "yaml"),
                    )
                except Exception as exc:
                    raise self.error(f"Error parsing data: {exc}")

        if "template" in self.options:
            # get template from file
            input_file = self.options["template"]
            # get file path relative to the document
            _, abs_path = self.env.relfn2path(input_file)
            if not Path(abs_path).exists():
                raise self.error(f"'template' path does not exist: {abs_path}")
            # read file
            with Path(abs_path).open() as handle:
                template_lines = handle.readlines()
            # add input file to dependencies
            self.env.note_dependency(input_file)

        # validate data
        if not isinstance(data, list):
            raise self.error("Data must be a list")
        if not data:
            raise self.error("Data must not be empty")

        for idx, item in enumerate(data):
            if not isinstance(item, dict):
                raise self.error(f"item {idx}: each data item must be a dictionary")

            if "start" not in item:
                raise self.error(f"item {idx}: each data item must contain 'start' key")
            try:
                item["start"] = dtime.to_datetime(item["start"])
            except Exception as exc:
                raise self.error(
                    f"item {idx}: error parsing 'start' value: {exc}"
                ) from exc

            if "duration" in item:
                try:
                    item["duration"] = dtime.parse_duration(item["duration"])
                except Exception as exc:
                    raise self.error(
                        f"item {idx}: error parsing 'duration' value: {exc}"
                    ) from exc

        # validate template
        if not [line for line in template_lines if line.strip()]:
            raise self.error("Template cannot be empty")

        env = jinja2.Environment()
        # env.filters["fdtime"] = _format_datetime
        template = env.from_string("\n".join(template_lines))

        container = TimelineDiv()
        if "height" in self.options:
            container.add_style("--tl-height", self.options["height"])
        if "width-item" in self.options:
            container.add_style("--tl-item-width", self.options["width-item"])

        list_node = nodes.enumerated_list(
            classes=[f"timeline-{self.options.get('style', 'default')}"]
            + self.options.get("class", [])
        )
        self.set_source_info(list_node)
        container.append(list_node)

        for idx, item in enumerate(
            sorted(
                data, key=lambda x: x["start"], reverse=("reversed" not in self.options)
            )
        ):
            if self.options.get("max-items") and idx >= self.options["max-items"]:
                break
            rendered = template.render(
                e=item,
                dt=dtime.DtRangeStr(item["start"]),
                duration=dtime.fmt_delta(item.get("duration")),
                dtrange=dtime.DtRangeStr(item["start"], item.get("duration")),
            )
            item_node = nodes.list_item(
                classes=(["timeline"] + self.options.get("class-item", []))
            )
            self.set_source_info(item_node)
            list_node.append(item_node)
            item_container = TimelineDiv(
                classes=["tl-item"], dt=item["start"].isoformat()
            )
            item_content = TimelineDiv(classes=["tl-item-content"])
            item_container.append(item_content)
            # item_container.append(nodes.Text(rendered))
            self.state.nested_parse(
                StringList(rendered.splitlines(), self.state.document.current_source),
                self.content_offset,
                item_content,
            )
            item_node.append(item_container)

        self.env.metadata[self.env.docname]["timeline"] = True

        return [container]


def read_events(
    stream: TextIO, fmt: Literal["yaml", "json", "csv"]
) -> list[dict[str, Any]]:
    """Read events from a stream."""
    if fmt == "yaml":
        return yaml.safe_load(stream)
    if fmt == "json":
        return json.load(stream)
    if fmt == "csv":
        return list(csv.DictReader(stream))

    raise ValueError(f"Unknown format: {fmt}")
